from random import randint, random

from games.screen import Screen
from games.objects import (Square, Explosion, Projectile, Bar, AbstractPlayer,
                           AbstractEnemies, CompassionateBoss)


class Player(AbstractPlayer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, score_title='Bashed', **kwargs)

    def reset(self):
        super().reset()
        self.continuous_moves = 0
        self.size = 3
        self.active = False
        if self.screen:
            self.x = int(self.screen.width / 2)
            self.y = self.screen.height - self.size

    def render(self, screen: Screen):
        super().render(screen)

        # Cap it
        move_cap = 100 if self.char == '^' else 5
        if self.continuous_moves > move_cap:
            self.continuous_moves = move_cap

        if self.continuous_moves % 30 == 0 and self.char == '^':
            self.size = max(1, 3 - self.continuous_moves / 30)

        if (self.screen.renders % 2 == 0 or self.char == '!') and self.active:
            if self.char == '!':
                color = self.screen.colors[int(self.screen.renders / 2) % len(self.screen.colors)]
            else:
                color = self.color

            if self.char == 'O':
                projectile = Bar(self.x+1, self.y-1, char=self.char, parent=self, color=color,
                                 y_delta=-1)
            else:
                projectile = Projectile(self.x, self.y-1, shape=self.char, parent=self, color=color)

            self.kids.add(projectile)
            self.screen.add(projectile)

    def destruct(self):
        super().destruct(msg='You got BASHED!!')

    def left_pressed(self):
        if self.continuous_moves >= 5:
            speed = 3 if self.char == '!' else 2
        else:
            speed = 1
        if self.active:
            if self.x - speed >= 1:
                self.x -= speed
                self.continuous_moves += 1
            elif self.x - 1 >= 1:
                self.x -= 1
                self.continuous_moves += 1

    def right_pressed(self):
        if self.continuous_moves >= 5:
            speed = 3 if self.char == '!' else 2
        else:
            speed = 1
        if self.active:
            if self.x + speed < self.screen.width - 1:
                self.x += speed
                self.continuous_moves += 1
            elif self.x + 1 < self.screen.width - 1:
                self.x += 1
                self.continuous_moves += 1

    def up_pressed(self):
        if self.active and self.y >= self.size:
            self.y -= 1
            self.continuous_moves += 1

    def down_pressed(self):
        if self.active and self.y < self.screen.height - self.size:
            self.y += 1
            self.continuous_moves += 1

    def key_released(self):
        if self.continuous_moves >= 1:
            self.continuous_moves -= 1


class Boss(CompassionateBoss):
    def got_hit(self):
        self.size = int(self.hp / self.max_hp * 3 + 2)

        if self.screen.renders % 10 == 0:
            if self.size >= 4:
                self.color = self.screen.COLOR_GREEN
            elif self.size >= 3:
                self.color = self.screen.COLOR_YELLOW
            else:
                self.color = self.screen.COLOR_RED

            return False
        else:
            return True

    def render(self, screen: Screen):
        super().render(screen)

        if not self.color:
            self.color = self.screen.COLOR_GREEN

        if self in screen:
            for projectile in self.kids:
                if projectile.coords & self.player.coords:
                    self.player.destruct()

            if self.player.size == 1 or not self.player.alive:
                self.x_delta = 0
            elif not self.x_delta:
                self.x_delta = self.initial_x_delta

            if self.screen.renders % 60 == 0 and self in self.screen:
                projectile = Bar(self.x, self.y+self.size/2, size=self.size,
                                 parent=self, color=self.color, y_delta=0.5,
                                 char=self.char)
                self.kids.add(projectile)
                self.screen.add(projectile)


class Enemies(AbstractEnemies):
    def create_enemy(self):
        return Square(randint(3, self.screen.width-3), -3, size=randint(2, 4),
                      y_delta=random() * self.player.score / 200 + 0.2)

    def on_death(self, enemy):
        """ Optionally add custom actions when an enemy dies """
        explosion_size = 25 if enemy == self.boss else enemy.size * 3
        self.screen.add(Explosion(enemy.x, enemy.y, size=explosion_size))

    def should_spawn_boss(self):
        """ Override this to customize logic for when a boss should be created """
        return self.player.score and self.player.score % 50 == 0

    def create_boss(self):
        return Boss('Max',
                    Square(randint(5, self.screen.width - 5), -5, size=5, char='$',
                           y_delta=0.1, solid=True, color=self.screen.COLOR_GREEN),
                    player=self.player,
                    hp=self.player.score)

    def additional_enemies(self):
        """ Increase enemies as the player levels up """
        return int(self.player.score / 100)



from games.manager import Manager
from games.controller import Controller
from games.screen import Scene


def test_manager():
    Manager()


def test_game(screen):
    class Game(Controller):
        name = 'Test'

        def init(self):
            self.scenes = [Scene]

    with screen:
        game = Game(screen)
        assert not game.done

        game.play()
        assert game.current_scene
        game.reset_scene()



from unittest.mock import Mock

import pytest

from games.controller import Controller
from games.screen import Screen, ScreenBuffer, Scene
from games.objects import Border, AbstractPlayer, Diamond


@pytest.fixture
def screen(monkeypatch):
    monkeypatch.setattr('curses.cbreak', Mock())
    monkeypatch.setattr('curses.nocbreak', Mock())
    monkeypatch.setattr('curses.endwin', Mock())

    screen = Screen(border=Border())
    screen._width = 80
    screen._height = 20
    screen.buffer = ScreenBuffer(80, 20)
    return screen


@pytest.fixture
def game(screen):
    class Game(Controller):
        name = 'Game Test'

        def init(self):
            self.scenes = [Scene]

    return Game(screen)


@pytest.fixture
def player(game):
    return AbstractPlayer('Jon', Diamond(40, 20), game)



from random import randint, random, choice

from games.screen import Screen
from games.objects import (Zombie, Explosion, Projectile, Monologue,
                           DyingZombie, AbstractPlayer, AbstractEnemies, CompassionateBoss)


class Player(AbstractPlayer):
    def __init__(self, *args, **kwargs):
        self.delta_index = 0
        self.ammos_limit = 1000
        self.gas_limit = 100
        self.deltas = ((0, -1, '|'), (1, -1, '/'), (1, 0, '-'), (1, 1, '\\'), (0, 1, '|'), (-1, 1, '/'),
                       (-1, 0, '-'), (-1, -1, '\\'))

        super().__init__(*args, score_title='Killed', **kwargs)

    def reset(self):
        super().reset()
        self.sync(self.shape)
        self.color = None
        self.delta_index = 0

        # Machine gun upgrade when score reaches 50
        self.ammos = self.ammos_limit
        self.using_machine_gun = False
        self.machine_gun_enabled = False

        # Grenade upgrade when score reaches 100
        self.grenades_limit = 10
        self.grenades = self.grenades_limit
        self.grenades_enabled = False

        # Machine gun upgrade when score reaches 50
        self.gas = self.gas_limit
        self.using_flamethrower = False
        self.flamethrower_enabled = False

        if self.screen:
            self.x = self.screen.width / 2
            self.y = self.screen.height / 2
            self.screen.status.pop('Ammos', None)
            self.screen.status.pop('Grenades', None)
            self.screen.status.pop('Gas', None)

    def render(self, screen: Screen):
        super().render(screen)

        if (self.screen.renders % 2 == 0 or self.using_machine_gun or self.using_flamethrower) and self.active:
            x_delta, y_delta, shape = self.deltas[self.delta_index]
            if self.using_machine_gun or self.using_flamethrower:
                x_delta *= 2
                y_delta *= 2
                color = screen.COLOR_YELLOW
                if self.using_flamethrower:
                    shape = None
            else:
                shape = chr(183)
                color = None
            if (not self.using_machine_gun or self.using_machine_gun and self.ammos > 0
                    or self.using_flamethrower and self.gas > 0):
                if self.using_flamethrower:
                    for flame_size in range(10):
                        explosion = Explosion(self.x, self.y, size=flame_size, parent=self)
                        projectile = Projectile(self.x, self.y, shape=shape, parent=self,
                                                x_delta=x_delta, y_delta=y_delta, color=color,
                                                explode_after_renders=flame_size,
                                                explosion=explosion)
                        self.kids.add(projectile)
                        self.screen.add(projectile)
                else:
                    projectile = Projectile(self.x, self.y, shape=shape, parent=self,
                                            x_delta=x_delta, y_delta=y_delta, color=color)
                    self.kids.add(projectile)
                    self.screen.add(projectile)

        if self.using_machine_gun:
            if self.ammos > 0:
                self.ammos -= 1
            else:
                self.using_machine_gun = False
                screen.add(Monologue(self.x, self.y - 1, texts=['Out of ammo!!!',
                                                                'Wait for reload before upgrading.']))
        elif self.ammos < self.ammos_limit:
            self.ammos += 1

        if self.using_flamethrower:
            if self.gas > 0:
                self.gas -= 0.5
            else:
                self.using_flamethrower = False
                self.using_machine_gun = True
                screen.add(Monologue(self.x, self.y - 1, texts=['Out of gas!!!',
                                                                'Wait for refill before upgrading.']))
        elif self.gas < self.gas_limit:
            self.gas += 0.1

        if self.grenades_enabled:
            self.grenades_limit = 10 + int((self.score - 100) / 10)
            if self.grenades < self.grenades_limit and self.screen.renders % 30 == 0:
                self.grenades += 1

        if self.score > 50 and not self.machine_gun_enabled:
            self.machine_gun_enabled = True
            self.using_machine_gun = True
            screen.add(Monologue(self.x, self.y - 1, ['New weapon unlocked: MACHINE GUN!!',
                                                      'It shoots and moves twice as fast,',
                                                      'but watch out as ammos are limited.',
                                                      'Press up/down to switch your weapon']))
            self.ammos = self.ammos_limit

        if self.score > 100 and not self.grenades_enabled:
            self.grenades_enabled = True
            screen.add(Monologue(self.x, self.y - 1, ['New weapon unlocked: GRENADES!!',
                                                      'You got {} of them.'.format(self.grenades_limit),
                                                      'It EXPLODES on impact.',
                                                      'Press Space to launch a grenade']))
            self.grenades = self.grenades_limit

        if self.score > 150 and not self.flamethrower_enabled:
            self.flamethrower_enabled = True
            self.using_flamethrower = True
            self.using_machine_gun = False
            screen.add(Monologue(self.x, self.y - 1, ['New weapon unlocked: FLAMETHROWER!!',
                                                      'Destroy everything before the gas runs out!',
                                                      'Press up/down to switch your weapon']))
            self.gas = self.gas_limit

        if self.machine_gun_enabled:
            screen.border.status['Ammos'] = self.ammos
        if self.grenades_enabled:
            screen.border.status['Grenades'] = self.grenades
            if self.alive:
                if self.grenades:
                    self.color = screen.COLOR_RAINBOW
                else:
                    self.color = None
        if self.flamethrower_enabled:
            screen.border.status['Gas'] = '{}%'.format(int(self.gas))

    def destruct(self):
        super().destruct(msg='You got ZOMBIFIED!!', explosion_size=30)
        self.screen.add(Zombie(self.shape.x, self.shape.y, color=self.screen.COLOR_GREEN, y_delta=-0.1))

    def left_pressed(self):
        if self.alive:
            if self.delta_index > 0:
                self.delta_index -= 1
            else:
                self.delta_index = len(self.deltas) - 1

    def right_pressed(self):
        if self.alive:
            if self.delta_index < len(self.deltas) - 1:
                self.delta_index += 1
            else:
                self.delta_index = 0

    def up_pressed(self):
        if self.using_machine_gun and self.flamethrower_enabled:
            self.using_flamethrower = True
            self.using_machine_gun = False
        elif self.machine_gun_enabled:
            self.using_machine_gun = True

    def down_pressed(self):
        if self.using_flamethrower:
            self.using_flamethrower = False
            self.using_machine_gun = True
        else:
            self.using_machine_gun = False

    def space_pressed(self):
        if self.grenades_enabled and self.grenades > 0 and self.alive:
            self.grenades -= 1
            x_delta, y_delta, shape = self.deltas[self.delta_index]
            explosion = Explosion(self.x, self.y, size=min(self.screen.width, self.screen.height),
                                  parent=self)
            projectile = Projectile(self.x, self.y, shape=chr(0x274d), parent=self,
                                    x_delta=x_delta, y_delta=y_delta, color=self.color,
                                    explode_after_renders=10,
                                    explosion=explosion, explosions=5)
            self.kids.add(projectile)
            self.screen.add(projectile)


class Enemies(AbstractEnemies):
    def create_enemy(self):
        if random() < 0.5:
            x = choice([0, self.screen.width])
            y = randint(0, self.screen.height)
        else:
            y = choice([0, self.screen.height])
            x = randint(0, self.screen.width)

        speed = random() * self.player.score / 10000 + 0.2
        x_sign = (1 if x < self.player.x else -1) * random()
        y_sign = (1 if y < self.player.y else -1) * random()

        return Zombie(x, y, x_delta=x_sign * speed, y_delta=y_sign * speed, random_start=True)

    def on_death(self, enemy):
        zombie = DyingZombie(enemy.x, enemy.y, color=self.screen.COLOR_RED)
        self.screen.add(zombie)

    def should_spawn_boss(self):
        return self.player.score and self.player.score % 50 == 0

    def create_boss(self):
        return CompassionateBoss('Max',
                                 Zombie(randint(5, self.screen.width - 5), -5,
                                        y_delta=0.1, color=self.screen.COLOR_GREEN, random_start=True),
                                 self.player,
                                 hp=self.player.score/25)



from random import randint, random

from games.screen import Screen
from games.listeners import KeyListener


class ScreenObject:
    """ Base class for all objects on screen """
    def __init__(self, x: int, y: int, x_delta=None, y_delta=None, color=None, size=1, parent=None):
        self.x = x
        self.y = y
        self.x_delta = x_delta
        self.y_delta = y_delta
        self.color = color
        self.size = size
        self.coords = set()
        self.parent = parent
        self.kids = set()
        self.screen = None
        self.visible = True

    def copy(self):
        obj = self.__class__(self.x, self.y, x_delta=self.x_delta, y_delta=self.y_delta,
                             color=self.color, size=self.size, parent=self.parent)
        obj.coords = self.coords
        obj.kids = self.kids
        obj.visible = self.visible
        obj.screen = self.screen
        return obj

    @property
    def all_kids(self):
        kids = self.kids.copy()
        all_kids = set()

        while kids:
            kid = kids.pop()
            all_kids.add(kid)

            if kid.kids:
                kids.update(kid.kids)

        return all_kids

    @property
    def all_coords(self):
        """ All coords of this object and its kids """
        coords = self.coords.copy()
        for kid in self.all_kids:
            coords.update(kid.coords)

        return coords

    @property
    def is_out(self):
        """ Indicates if the object center is outside of the screen border """
        try:
            return (self.x + self.size < 0 or self.x - self.size > self.screen.width
                    or self.y + self.size < 0 or self.y - self.size > self.screen.height)

        except Exception:
            return False

    def sync(self, screen_object, location_only=False):
        self.x = screen_object.x
        self.y = screen_object.y

        if not location_only:
            self.x_delta = screen_object.x_delta
            self.y_delta = screen_object.y_delta
            self.size = screen_object.size
            self.color = screen_object.color
            self.visible = screen_object.visible
            self.coords = screen_object.coords

    def render(self, screen: Screen):
        """ Render object onto the given screen """
        self.screen = screen

        if self.x_delta:
            self.x += self.x_delta

        if self.y_delta:
            self.y += self.y_delta

    def reset(self):
        """ Reset object to original state """
        if self.screen:
            for kid in self.kids:
                if kid in self.screen:
                    self.screen.remove(kid)
        self.kids = set()

    def add_kid(self, screen_object):
        self.kids.add(screen_object)

    def remove_kid(self, screen_object):
        if screen_object in self.kids:
            self.kids.remove(screen_object)

    def replace_kid(self, old_object, new_object):
        self.remove_kid(old_object)
        self.add_kid(new_object)


class ScreenObjectGroup(ScreenObject):
    """ Group of objects that move together as one """

    def __init__(self, *args, objects=None, add_bar=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.ordered_objects = list(objects) if objects else []
        self.add_bar = add_bar

    def add(self, *screen_objects):
        self.ordered_objects.extend(screen_objects)

    def render(self, screen):
        super().render(screen)

        if self.kids:
            for kid in self.kids:
                kid.y = int(self.y + 3 if isinstance(kid, Bar) else self.y)

        else:
            padding = 1
            width = sum((o.size + padding) for o in self.ordered_objects)
            block_size = width / len(self.ordered_objects)
            start_x = int(self.x - width / 2) - 2

            for obj in self.ordered_objects:
                start_x += block_size
                obj.x = start_x

                screen.add(obj)
                self.kids.add(obj)

            if self.add_bar:
                divider = Bar(self.x + 1, self.y + 3, size=width + 1)
                screen.add(divider)
                self.kids.add(divider)


class AbstractPlayer(ScreenObject, KeyListener):
    def __init__(self, name, shape: ScreenObject, controller, score_title='score', show_total=False,
                 max_hp=1):
        super().__init__(shape.x, shape.y, size=shape.size, color=shape.color)

        self.name = name
        self.shape = shape
        self._original_shape = shape
        self.high_score = 0
        self.total_score = 0
        self.show_total = show_total
        self.char = shape.char
        self.controller = controller
        self.score_title = score_title
        self.max_hp = max_hp

        self.reset()

    @property
    def visible(self):
        return self.alive

    @visible.setter
    def visible(self, visible):
        self.alive = visible

    def reset(self):
        super().reset()
        self.hp = self.max_hp
        self.score = 0
        self.alive = True
        self.active = True  # Indicates if the player is taking action, such as shooting
        self.is_hit = False
        self.size = self.shape.size
        self.shape = self._original_shape
        if self.screen:
            self.x = int(self.screen.width / 2) + 1
            self.y = self.screen.height - int(self.size / 2) - 1

    def render(self, screen: Screen):
        super().render(screen)

        # screen.debug(kids=len(self.kids))

        self.shape.sync(self)
        self.shape.render(screen)
        self.coords = self.shape.coords

        score = str(self.score)
        if self.score < self.high_score:
            score += ' | High: {}'.format(self.high_score)
        if self.show_total and self.score < self.total_score:
            score += ' | Total: {}'.format(self.total_score)

        self.screen.status[self.score_title] = score

    def scored(self, points=1):
        """ Add given points, defaults to 1, to the player's score """
        self.score += points
        self.total_score += points
        if self.score > self.high_score:
            self.high_score = self.score

    def got_hit(self, points=1):
        """ Decrease player's HP by the given points and destruct when HP hits 0 """
        self.is_hit = True
        self.hp -= points
        if self.hp <= 0:
            self.hp = 0
            self.destruct()

    def destruct(self, msg='The End', explode=True, explosion_size=20):
        self.alive = False
        self.active = False
        self.screen.add(Text(self.screen.width / 2, self.screen.height / 2, msg,
                             is_centered=True))
        if explode:
            self.screen.add(Explosion(self.x, self.y, size=explosion_size,
                                      on_finish=self.controller.reset_scene))


class CompassionateBoss(ScreenObject):
    def __init__(self, name, shape: ScreenObject, player: AbstractPlayer, hp=5):
        super().__init__(shape.x, shape.y, size=shape.size, color=shape.color,
                         x_delta=shape.x_delta, y_delta=shape.y_delta)
        self.name = name
        self.shape = shape
        self.player = player
        if self.y_delta is None:
            self.y_delta = 0.1
        if self.x_delta is None:
            self.x_delta = self.initial_x_delta
        self.is_hit = False
        self.char = shape.char
        self.hp = hp
        self.max_hp = hp

    @property
    def initial_x_delta(self):
        return max(random() * 0.5, 0.2)

    def got_hit(self):
        """ React to getting hit and return state for is_hit """
        return False

    def render(self, screen: Screen):
        super().render(screen)
        self.shape.sync(self)
        self.shape.render(screen)
        self.coords = self.shape.coords

        if self in self.screen:
            if self.x > self.player.x:
                self.x_delta = -1 * abs(self.x_delta)
            else:
                self.x_delta = abs(self.x_delta)

            if self.y > self.screen.height + 2.5 and self.player.alive:
                self.y = -2.5

            if self.is_hit:
                self.color = self.screen.colors[self.screen.renders % len(self.screen.colors)]
                self.is_hit = self.got_hit()


class AbstractEnemies(ScreenObject):
    def __init__(self, player: AbstractPlayer, max_enemies=5):
        super().__init__(0, 0)
        self.max_enemies = max_enemies
        self.enemies = set()
        self.player = player
        self.boss = None

    def create_enemy(self) -> ScreenObject:
        """ Create an enemy instance """
        raise NotImplementedError

    def on_death(self, enemy):
        """ Optionally add custom actions when an enemy dies """

    def should_spawn_boss(self):
        """ Override this to customize logic for when a boss should be created """
        return False

    def create_boss(self) -> ScreenObject:
        """ Create a boss instance """
        raise NotImplementedError

    def additional_enemies(self):
        """ Increase enemies as the player levels up """
        return int(self.player.score / 100)

    def render(self, screen: Screen):
        super().render(screen)

        if self.player.alive:
            # Create enemies
            if len(self.enemies) < self.max_enemies + self.additional_enemies():
                enemy = self.create_enemy()
                self.enemies.add(enemy)
                screen.add(enemy)

            # Create boss
            if self.should_spawn_boss() and self.boss not in screen:
                self.boss = self.create_boss()
                self.enemies.add(self.boss)
                screen.add(self.boss)

        for enemy in list(self.enemies):
            # Make them go fast when player is destroyed
            if not self.player.alive:
                if enemy.y_delta:
                    enemy.y_delta *= 1.2
                if enemy.x_delta:
                    enemy.x_delta *= 1.2

            # If it is out of the screen, remove it (except for boss or player is dead)
            if enemy.is_out and (enemy != self.boss or not self.player.alive):
                self.enemies.remove(enemy)
                screen.remove(enemy)

            # Otherwise, check if player's projectiles hit the enemies
            else:
                for projectile in list(self.player.kids):
                    if projectile.coords & enemy.coords:
                        if enemy == self.boss and self.boss.hp > 0:
                            self.boss.hp -= 1
                            self.boss.is_hit = True
                            break

                        self.enemies.remove(enemy)
                        screen.remove(enemy)

                        self.player.kids.remove(projectile)
                        screen.remove(projectile)

                        self.on_death(enemy)

                        if hasattr(projectile, 'explode'):
                            projectile.explode()

                        self.player.scored(points=5 if enemy == self.boss else 1)

                        break
                else:
                    if enemy.all_coords & self.player.coords and self.player.alive:
                        self.player.got_hit()


class Bitmap(ScreenObject):
    def __init__(self, *args, char=None, random_start=False, remove_after_animation=False,
                 flip=False, centered=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.char = getattr(self, 'char', char)  # or chr(0x2588)
        self._bitmaps = getattr(self, 'bitmaps', [])
        self._frames_per_bitmap = getattr(self, 'frames_per_bitmap', 10)
        self._remove_after_animation = getattr(self, 'remove_after_animation', remove_after_animation)
        self._bitmap_index_offset = randint(0, max(len(self._bitmaps), 1) - 1) if random_start else 0
        self.renders = 0
        self.flip = flip
        self.centered = centered
        self._bitmap = self._bitmaps and self._bitmaps[0] or getattr(self, 'bitmap', """\
▓▓▓▓▓
▓▓▓▓▓
▓▓▓▓▓
▓▓▓▓▓
▓▓▓▓▓
""")  # noqa
        self.size = len(self._bitmap.strip('\n').split('\n'))

    def draw(self, x, y, char, screen: Screen):
        screen.draw(x, y, char, color=self.color)
        self.coords.add((x, y))

    def render(self, screen: Screen):
        super().render(screen)
        self.renders += 1

        if self._bitmaps:
            index = (self._bitmap_index_offset + int(self.renders / self._frames_per_bitmap)) % len(self._bitmaps)
            bitmap = self._bitmaps[index]
        else:
            bitmap = self._bitmap
        bitmap = bitmap.strip('\n').split('\n')

        x_size = len(bitmap[0])
        if self.centered:
            start_x = int(self.x - x_size / 2)
            start_y = int(self.y - self.size / 2)
        else:
            start_x = int(self.x)
            start_y = int(self.y)
        self.coords = set()

        for y in range(start_y, start_y + self.size):
            x_size = len(bitmap[y-start_y])
            for x in range(start_x, start_x + x_size):
                if self.flip:
                    x_offset = (start_x + x_size - 1) - x
                else:
                    x_offset = x - start_x
                if bitmap[y-start_y][x_offset] != ' ':
                    self.draw(x, y, self.char or bitmap[y-start_y][x_offset], screen)

        if self._remove_after_animation and self.renders > self._frames_per_bitmap * (len(self._bitmaps) - 1):
            screen.remove(self)
            if self.parent:
                self.parent.remove_kid(self)


class ObjectMap(Bitmap):
    """ Represents screen objects using characters on a bitmap """
    def __init__(self, *args, grid_size=5, **kwargs):
        super().__init__(*args, centered=False, **kwargs)

        #: Number of characters each object represents
        self.grid_size = grid_size

        #: Map of object characters to colors
        self._object_map = getattr(self, 'object_map', {})

    @property
    def is_out(self):
        return False

    def draw(self, x, y, char, screen):
        x = (x - int(self.x)) * self.grid_size + int(self.x)
        y = (y - int(self.y)) * self.grid_size + int(self.y) + self.grid_size / 2
        if char in self.object_map:
            obj_cls, color_name = self.object_map[char]
            obj = obj_cls(x, y, color=color_name and getattr(screen, color_name))
            if not obj.is_out:
                obj.renders = self.renders + x
                obj.render(screen)
                self.coords.update(obj.coords)


class Text(ScreenObject):
    def __init__(self, x: int, y: int, text: str, x_delta=None, y_delta=None, is_centered=False, color=None):
        super().__init__(x, y, size=int(len(text) / 2), x_delta=x_delta, y_delta=y_delta, color=color)
        self.text = text
        self.is_centered = is_centered

    def render(self, screen: Screen):
        super().render(screen)

        if self.text:
            self.coords = set()
            x = self.x - len(self.text) / 2 if self.is_centered else self.x
            for offset, char in enumerate(self.text):
                screen.draw(x + offset, self.y, char, color=self.color)
                self.coords.add((int(x + offset), int(self.y)))


class Logo(ScreenObject):
    def __init__(self, x, y, name: str, shape: ScreenObject, color=None):
        super().__init__(x, y, size=7, color=color)
        self.text = Text(self.x, self.y + 3, name, is_centered=True, color=self.color)
        self.shape = shape

    def render(self, screen: Screen):
        super().render(screen)

        self.text.sync(self, location_only=True)
        self.text.y += 3
        self.shape.sync(self, location_only=True)

        self.text.render(screen)
        self.shape.render(screen)


class Monologue(Text):
    def __init__(self, x: int, y: int, texts=[], on_finish=None, x_delta=None, y_delta=None):
        super().__init__(x, y, texts[0], x_delta=x_delta, y_delta=y_delta)
        self.center_x = x
        self.texts = texts
        self.on_finish = on_finish
        self.reset()

    def reset(self):
        super().reset()
        self.index = 0
        self.renders = 0

    def render(self, screen: Screen):
        self.renders += 1
        self.text = self.texts[self.index]
        self.x = self.center_x - len(self.text) / 2

        super().render(screen)

        if self.renders % 45 == 0:
            self.index += 1
            if self.index >= len(self.texts):
                screen.remove(self)
                if self.on_finish:
                    self.on_finish()


class Border(ScreenObject):
    def __init__(self, char: str = None, show_fps=False, title=None):
        super().__init__(0, 0)
        self.char = char
        self.show_fps = show_fps
        self.title = title
        self.reset()

    def reset(self):
        super().reset()

        #: A value from 0 to 1 indicating the level of player's health
        self.health_level = 0

        #: A value from 0 to 1 indicating the level of player's energy
        self.energy_level = 0
        self.status = {}

    def set_levels(self, health_level, energy_level):
        health_level = round(health_level, 2)  # Easier to compare with
        energy_level = round(energy_level, 2)
        if self.health_level != health_level:
            self.health_level = health_level

        if self.energy_level != energy_level:
            self.energy_level = energy_level

    def render(self, screen: Screen):
        for x in range(screen.width):
            for y in range(screen.height):
                if x == 0 or y == 0 or x == screen.width - 1 or y == screen.height - 1:
                    char = (self.char
                            or (x == 0 and y == 0) and chr(0x2554)
                            or (x == screen.width - 1 and y == 0) and chr(0x2557)
                            or (x == 0 and y == screen.height - 1) and chr(0x255A)
                            or (x == screen.width - 1 and y == screen.height - 1) and chr(0x255D)
                            or (y == 0 or y == screen.height - 1) and chr(0x2550)
                            or (x == 0 or x == screen.width - 1) and chr(0x2551))

                    if self.health_level and x == 0 and y < screen.height - 1:
                        y_level = int((screen.height - 1) * self.health_level + 0.5)
                        if screen.height - 1 - y <= y_level:
                            screen.draw(1, y, '│', color=screen.COLOR_RED)

                    if self.energy_level and x == screen.width - 1 and y < screen.height - 1:
                        y_level = int((screen.height - 2) * self.energy_level + 0.5)
                        if screen.height - 1 - y <= y_level:
                            screen.draw(screen.width - 2, y, '│', color=screen.COLOR_GREEN)

                    screen.draw(x, y, char, color=self.color)

        if self.title:
            padded_title = ' ' + self.title + ' '
            start_x = round((screen.width - len(padded_title)) / 2)
            for x_offset in range(len(padded_title)):
                screen.draw(start_x + x_offset, 0, padded_title[x_offset], color=self.color)

        if self.show_fps and screen.fps:
            self.status['FPS'] = screen.fps

        if self.status:
            debug_text = ' ' + ' | '.join(
                ['{}: {}'.format(k[0].upper() + k[1:], v) for k, v in self.status.items()]) + ' '
            start_x = round((screen.width - len(debug_text)) / 2)
            for x_offset in range(len(debug_text)):
                screen.draw(start_x + x_offset, screen.height - 1, debug_text[x_offset], color=self.color)


class Char(ScreenObject):
    def __init__(self, *args, char, **kwargs):
        super().__init__(*args, **kwargs)
        self.char = char

    def render(self, screen: Screen):
        super().render(screen)

        self.coords = {(int(self.x), int(self.y))}

        if not self.is_out:
            screen.draw(self.x, self.y, self.char, color=self.color)


class Projectile(ScreenObject):
    def __init__(self, x: int, y: int, shape='^', x_delta=None, y_delta=-1, color=None, size=1,
                 parent=None, explode_after_renders=None, explosion=None, explosions=1):
        super().__init__(x, y, color=color, x_delta=x_delta, y_delta=y_delta, size=size, parent=parent)
        self.shape = shape
        self.explode_after_renders = explode_after_renders
        self.renders = 0
        self.explosion = explosion
        self.explosions = explosions

    @property
    def char(self):
        return self.shape

    def render(self, screen: Screen):
        super().render(screen)

        self.renders += 1

        self.coords = set() if self.shape is None else {(int(self.x), int(self.y))}

        if not self.is_out and self.shape is not None:
            screen.draw(self.x, self.y, self.shape, color=self.color)

        if (self.explode_after_renders and self.renders >= self.explode_after_renders):
            self.explode()

    def explode(self):
        if self.explosion:
            if self.explosions:
                explosion = self.explosion.copy()
                self.x_delta = 0
                self.y_delta = 0
                explosion.x = self.x
                explosion.y = self.y
                self.screen.add(explosion)
                self.parent.add_kid(explosion)
                self.explosions -= 1

            else:
                self.screen.remove(self)
                self.parent.remove_kid(self)


class Circle(ScreenObject):
    def __init__(self, x: int, y: int, size=5, char='O', x_delta=None, y_delta=None, color=None,
                 name=None):
        super().__init__(x, y, x_delta=x_delta, y_delta=y_delta, color=color,
                         size=size)
        self.char = char
        self.name = name

    def render(self, screen: Screen):
        super().render(screen)

        self.coords = {(int(self.x - 1), int(self.y - 1)), (int(self.x), int(self.y - 1)),
                       (int(self.x + 1), int(self.y - 1)), (int(self.x - 2), int(self.y)),
                       (int(self.x + 2), int(self.y)), (int(self.x - 1), int(self.y + 1)),
                       (int(self.x), int(self.y + 1)), (int(self.x + 1), int(self.y + 1))}

        for x, y in self.coords:
            screen.draw(x, y, self.char, color=self.color)


class Diamond(ScreenObject):
    def __init__(self, x: int, y: int, size=2, char='!', x_delta=None, y_delta=None, color=None,
                 name=None):
        super().__init__(x, y, x_delta=x_delta, y_delta=y_delta, color=color,
                         size=size)
        self.char = char
        self.name = name

    def render(self, screen: Screen):
        super().render(screen)

        self.coords = {(int(self.x + 1), int(self.y)), (int(self.x - 1), int(self.y)),
                       (int(self.x), int(self.y + 1)), (int(self.x), int(self.y - 1))}

        for x, y in self.coords:
            screen.draw(x, y, self.char, color=self.color)


class Square(ScreenObject):
    def __init__(self, x: int, y: int, size=3, char='#', x_delta=None, y_delta=None, color=None,
                 name=None, solid=False):
        super().__init__(x, y, x_delta=x_delta, y_delta=y_delta, color=color,
                         size=size)
        self.char = char
        self.name = name
        self.solid = solid

    def render(self, screen: Screen):
        super().render(screen)

        start_x = int(self.x - self.size / 2)
        start_y = int(self.y - self.size / 2)
        self.coords = set()

        for x in range(start_x, int(start_x + self.size)):
            for y in range(start_y, int(start_y + self.size)):
                if self.solid or (x == start_x or x == start_x + self.size - 1
                                  or y == start_y or y == start_y + self.size - 1):
                    self.coords.add((int(x), int(y)))
                    screen.draw(x, y, self.char, color=self.color)


class Explosion(ScreenObject):
    def __init__(self, x: int, y: int, size=10, char='*', on_finish=None, **kwargs):
        super().__init__(x, y, size=size, **kwargs)

        self.current_size = 2
        self.char = char
        self.on_finish = on_finish

    def copy(self):
        obj = super().copy()
        obj.current_size = self.current_size
        obj.char = self.char
        obj.on_finish = self.on_finish
        return obj

    def render(self, screen: Screen):
        if self.current_size > self.size:
            screen.remove(self)
            if self.parent:
                self.parent.remove_kid(self)
            if self.on_finish:
                self.on_finish()
            return

        super().render(screen)

        start_x = int(self.x - self.current_size / 2)
        start_y = int(self.y - self.current_size / 2)
        self.coords = set()
        colors = [screen.COLOR_YELLOW, screen.COLOR_RED]

        for x in range(start_x, start_x + self.current_size):
            for y in range(start_y, start_y + self.current_size):
                if (x == start_x or x == start_x + self.current_size - 1
                        or y == start_y or y == start_y + self.current_size - 1):
                    distance = ((x - self.x) ** 2 + (y - self.y) ** 2) ** (1/2)
                    if distance < self.current_size and random() < 2/self.current_size:
                        self.coords.add((int(x), int(y)))
                        screen.draw(x, y, self.char, color=colors[randint(0, len(colors) - 1)])

        self.current_size += 1


class Triangle(ScreenObject):
    def __init__(self, x: int, y: int, size=5, char='^', x_delta=None, y_delta=None, color=None,
                 name=None):
        super().__init__(x, y, x_delta=x_delta, y_delta=y_delta, color=color,
                         size=size)
        self.char = char
        self.name = name

    def render(self, screen: Screen):
        super().render(screen)

        self.coords = {(int(self.x), int(self.y - 1))}

        if self.size >= 2:
            self.coords.update({(int(self.x - 1), int(self.y)), (int(self.x + 1), int(self.y))})
            if self.size == 2:
                self.coords.update({(int(self.x), int(self.y))})

        if self.size >= 3:
            self.coords.update({(int(self.x - 1), int(self.y + 1)), (int(self.x), int(self.y + 1)),
                                (int(self.x + 1), int(self.y + 1)), (int(self.x - 2), int(self.y + 1)),
                                (int(self.x + 2), int(self.y + 1))})

        for x, y in self.coords:
            screen.draw(x, y, self.char, color=self.color)


class Bar(ScreenObject):
    def __init__(self, x: int, y: int, size=3, char='=', x_delta=None, y_delta=None, color=None,
                 parent=None):
        super().__init__(x, y, x_delta=x_delta, y_delta=y_delta, color=color,
                         size=size, parent=parent)
        self.char = char

    def render(self, screen: Screen):
        super().render(screen)

        start_x = int(self.x - self.size / 2 - 0.5)  # Round down
        self.coords = set()

        for x in range(start_x, int(start_x + self.size)):
            self.coords.add((int(x), int(self.y)))
            screen.draw(x, self.y, self.char, color=self.color)


class Choice(ScreenObject, KeyListener):
    def __init__(self, x: int, y: int, color=None, choices=[], on_select=None, on_change=None, current=None):
        super().__init__(x, y, color=color)

        self.choices = choices
        self.on_select = on_select
        self.on_change = on_change
        self._current = current if current is not None else int(len(self.choices) / 2 - 0.5)

        self.reset()

    @property
    def current(self):
        return self.choices[self._current]

    def reset(self):
        super().reset()
        self.bar = None

    def render(self, screen: Screen):
        super().render(screen)

        if not self.bar:
            max_size = max(c.size for c in self.choices)
            bar_size = max_size * 4
            width = bar_size * len(self.choices)
            start_x = int(self.x - width / 2)
            end_x = start_x + width

            self.bar = Bar(self.x, self.y + max_size / 2 + 1, size=bar_size, color=self.color or screen.COLOR_CYAN)
            screen.add(self.bar)
            self.kids.add(self.bar)

            instruction = Monologue(self.x, self.y + 1.5 * max_size,
                                    texts=['Change selection using arrow keys',
                                           'Press Space to start or Esc to exit'])
            screen.add(instruction)
            self.kids.add(instruction)

            for i, x in enumerate(range(start_x, end_x, bar_size)):
                choice = self.choices[i]
                choice.x = x + bar_size / 2
                choice.y = self.y
                screen.add(choice)
                self.kids.add(choice)

                if i == self._current:
                    self.bar.x = choice.x + 1

    def left_pressed(self):
        if self._current > 0 and self.bar:
            self._current -= 1
            self.bar.x -= self.bar.size
            if self.on_change:
                self.on_change(self.choices[self._current])

    def right_pressed(self):
        if self._current < len(self.choices) - 1 and self.bar:
            self._current += 1
            self.bar.x += self.bar.size
            if self.on_change:
                self.on_change(self.choices[self._current])

    def enter_pressed(self):
        if self.on_select:
            self.on_select(self.choices[self._current])

    def space_pressed(self):
        self.enter_pressed()


class One(Bitmap):
    represents = '1'
    bitmap = """
 ▓▓  
  ▓  
  ▓  
  ▓  
 ▓▓▓ 
"""  # noqa


class Two(Bitmap):
    represents = '2'
    bitmap = """
 ▓▓▓ 
▓   ▓
   ▓ 
 ▓▓  
▓▓▓▓▓
"""  # noqa


class Three(Bitmap):
    represents = '3'
    bitmap = """
▓▓▓▓ 
    ▓
 ▓▓▓ 
    ▓
▓▓▓▓ 
"""  # noqa


class Four(Bitmap):
    represents = '4'
    bitmap = """\
  ▓▓ 
 ▓ ▓ 
▓  ▓ 
▓▓▓▓▓
   ▓ 
"""  # noqa


class Five(Bitmap):
    represents = '5'
    bitmap = """\
▓▓▓▓▓
▓    
▓▓▓▓ 
    ▓
▓▓▓▓ 
"""  # noqa


class Six(Bitmap):
    represents = '6'
    bitmap = """\
 ▓▓▓ 
▓    
▓▓▓▓ 
▓   ▓
 ▓▓▓ 
"""  # noqa


class Seven(Bitmap):
    represents = '7'
    bitmap = """\
▓▓▓▓▓
   ▓ 
  ▓  
 ▓   
▓    
"""  # noqa


class Eight(Bitmap):
    represents = '8'
    bitmap = """\
 ▓▓▓ 
▓   ▓
 ▓▓▓ 
▓   ▓
 ▓▓▓ 
"""  # noqa


class Nine(Bitmap):
    represents = '9'
    bitmap = """\
 ▓▓▓ 
▓   ▓
 ▓▓▓▓
    ▓
 ▓▓▓ 
"""  # noqa


class Zero(Bitmap):
    represents = '0'
    bitmap = """\
 ▓▓▓ 
▓   ▓
▓   ▓
▓   ▓
 ▓▓▓ 
"""  # noqa


class Plus(Bitmap):
    represents = '+'
    bitmap = """\
     
  │  
 ─┼─ 
  │  
     
"""  # noqa


class Minus(Bitmap):
    represents = '-'
    bitmap = """\
     
     
 ─── 
     
     
"""  # noqa


class Multiply(Bitmap):
    represents = '*'
    bitmap = """\
     
 ╲ ╱ 
  ╳  
 ╱ ╲ 
     
"""  # noqa


class Divide(Bitmap):
    represents = '/'
    bitmap = """\
     
   ╱ 
  ╱  
 ╱   
     
"""  # noqa


class Space(Bitmap):
    represents = ' '
    bitmap = """
     
     
     
     
     
"""  # noqa


class Zombie(Bitmap):
    bitmaps = (r"""
   O 
 \-/\
  / |
 /\  
/ |  
""",  # noqa
r"""
  O  
\-/\ 
  | |
 /\  
/ |  
""",  # noqa
r"""
  O  
 -|\_
/ |  
 /\  
 | \ 
""",  # noqa
r"""
  O  
 -|\|
/ |  
 /\  
 | \ 
""",  # noqa
r"""
  O  
 -|\_
/ /  
 /\  
/  \ 
""") # noqa


class DyingZombie(Bitmap):
    remove_after_animation = True
    frames_per_bitmap = 3
    bitmaps = (r"""
   O 
 \-/\
  / |
 /\  
/ |  
""",  # noqa
r"""
     
  O  
\-/\ 
 /\  
/ |  
""",  # noqa
r"""
     
     
 O   
\-|\_
_| \_
""",  # noqa
r"""
     
     
     
| O\ 
\-|\|
""",  # noqa
r"""
     
     
     
_  \ 
\-O-/
""",  # noqa
r"""
     
     
     
     
\-O-/
""") # noqa


class Stickman(Bitmap):
    bitmap = r"""
 ☻ 
/|\
/ \
"""  # noqa


class StickmanWorried(Bitmap):
    bitmap = r"""
 ☺ 
\|/
/ \
"""  # noqa


class StickmanScared(Bitmap):
    bitmap = r"""
\☹/
 | 
/ \
"""  # noqa


class Wasp(Bitmap):
    frames_per_bitmap = 1
    bitmaps = (r"""
 vv
8oQ
""",  # noqa
r"""
 ww
8oQ
""")  # noqa noqa


class WaspKaiju(Bitmap):
    bitmaps = (r"""
 __      __
/  \_--_/  \
   |O  O|
    \  /
     \/
""",  # noqa
r"""
--_      __
   \_--_/  \
   |O  O|
    \  /
     ||
""",  # noqa
r"""
-__      ___
   \_--_/
   |O  O|
    \  /
     \/
""",  # noqa
r"""
--_      __
   \_--_/  \
   |O  O|
    \  /
     ||
""")  # noqa


class DyingWaspKaiju(Bitmap):
    remove_after_animation = True
    frames_per_bitmap = 3
    bitmaps = (r"""
 __      __
/  \_--_/  \
   \O  O/
    |  |
    \\//
""",  # noqa noqa
r"""
 
--_      _--
   \_--_/
   \O  O/
    \||/
""",  # noqa noqa
r"""
 
 
__      __
  \O  O/
   \\//
""",  # noqa noqa
r"""
 
 
 
_     _
 \O-O/
""") # noqa


class HealthPotion(Char):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, char='♥', **kwargs)


class Tree(Bitmap):
    bitmaps = (r"""
 __  /   __
//\\|\\_/\\\
// |\||/
     ||
     ||
     ||
""",  # noqa
r"""
 __  |   __
\//\|\\_/\ /
\/ |\||\
     ||
     ||
     ||
""",  # noqa
r"""
 __  \   __
 \/\|\\_/\ \
\\ |\||\
     ||
     ||
     ||
""")  # noqa


class Landscape(ObjectMap, KeyListener):
    move_speed = 0
    object_map = {
        'S': (Circle, 'COLOR_YELLOW'),  # Sun
        'T': (Tree, 'COLOR_GREEN')
    }

    def left_pressed(self):
        self.x += self.move_speed

    def right_pressed(self):
        self.x -= self.move_speed



from games.screen import Scene
from games.objects import Circle, Monologue, Triangle, Diamond, Choice
from games.geo_bash.objects import Player, Enemies


class ChoosePlayer(Scene):
    def init(self):
        choices = [
            Player('Kate', Triangle(int(self.screen.width / 2), self.screen.height - 3, size=3,
                                    color=self.screen.COLOR_BLUE), self.controller),
            Player('Nina', Circle(int(self.screen.width / 2), self.screen.height - 3, size=3,
                                  color=self.screen.COLOR_RED), self.controller),
            Player('Jon', Diamond(int(self.screen.width / 2), self.screen.height - 3, size=3,
                                  color=self.screen.COLOR_YELLOW), self.controller)
        ]
        self.choice = Choice(self.screen.width / 2, self.screen.height / 2,
                             choices=choices, color=self.screen.COLOR_CYAN,
                             on_select=self.next,
                             current=getattr(self.controller, '_last_player_index', None))

    def start(self):
        self.screen.add(self.choice)

    def next(self, chosen: Player):
        super().next()
        chosen.y = self.screen.height - chosen.size
        chosen.x = self.screen.width / 2
        self.controller.player = chosen
        self.controller._last_player_index = self.choice.choices.index(chosen)
        chosen.controller = self.controller

    def escape_pressed(self):
        self.controller.done = True


class Intro(Scene):
    def init(self):
        if self.controller.player.name == 'Jon':
            intro = 'the fastest shape'
        elif self.controller.player.name == 'Kate':
            intro = 'the stealthiest shape'
        else:
            intro = 'the most powerful shape'
        self.intro = Monologue(self.controller.player.x, self.controller.player.y - 2,
                               on_finish=self.next,
                               texts=["Hi, I am {} -- {}!".format(self.controller.player.name, intro),
                                      "Most people don't like geometry, but not me.",
                                      "I LOVE to bash them!! :D",
                                      "Move me using arrow keys.",
                                      "Ready? Let's BASH!!"])
        self.controller.player.reset()

    def start(self):
        self.screen.add(self.intro, self.controller.player)

    def key_pressed(self, key):
        self.next()


class Bash(Scene):
    def init(self):
        self.enemies = Enemies(self.controller.player)
        self.player.reset()
        self.controller.player.active = True

    def start(self):
        self.screen.add(self.player, self.enemies)

    def escape_pressed(self):
        self.next()



from games.screen import Scene
from games.objects import Monologue
from games.last_survivor.objects import Enemies


class Intro(Scene):
    def init(self):
        self.intro = Monologue(self.controller.player.x, self.controller.player.y - 2,
                               on_finish=self.next,
                               texts=["Hi, I am {}, THE LAST SURVIVOR".format(self.controller.player.name),
                                      "of the zombie apocalypse.",
                                      "It all started by a crazy scientist",
                                      "experimenting with a rat",
                                      "and injected it with the wrong formula.",
                                      "It turned into a zombie ...",
                                      "and spreaded to everyone.",
                                      "Oh no! Here they come. GET READY!!",
                                      "Move your weapon using left/right keys."])
        self.controller.player.reset()
        self.controller.player.active = False

    def start(self):
        self.screen.add(self.intro, self.controller.player)

    def key_pressed(self, key):
        self.next()


class Survive(Scene):
    def init(self):
        self.enemies = Enemies(self.controller.player)
        self.player.reset()

    def start(self):
        self.screen.add(self.player, self.enemies)

    def escape_pressed(self):
        self.controller.done = True



from random import randint, choice

from games.screen import Screen
from games.objects import (ScreenObject, KeyListener, Explosion, Text, ScreenObjectGroup, Bitmap,
                           One, Two, Three, Four, Five, Six, Seven, Eight, Nine, Zero,
                           Plus, Minus, Multiply, Divide, Space, Stickman, StickmanScared,
                           StickmanWorried, AbstractPlayer)


class Player(AbstractPlayer):
    def __init__(self, controller):
        super().__init__('Jon', Stickman(int(controller.screen.width / 2) + 1, controller.screen.height - 2),
                         controller, score_title='Crushed', show_total=True)

    def react(self, formula: ScreenObject):
        distance = self.y - formula.y
        if distance < 10:
            self.scared()
        elif distance < 15:
            self.worried()
        else:
            self.smile()

    def scared(self):
        self.shape = StickmanScared(self.x, self.y)

    def worried(self):
        self.shape = StickmanWorried(self.x, self.y)

    def smile(self):
        self.shape = Stickman(self.x, self.y)


class Formula(ScreenObjectGroup):
    BITMAPS = dict((b.represents, b) for b in [Zero, One, Two, Three, Four, Five, Six, Seven, Eight,
                                               Nine, Plus, Minus, Multiply, Divide, Space])

    def __init__(self, *args, text, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text
        for char in text:
            self.add(self.BITMAPS[char](self.x, self.y))


class Numbers(ScreenObject, KeyListener):
    def __init__(self, x, y, player: Player):
        super().__init__(x, y)
        self.player = player
        self.a = None
        self.b = None
        self.formula = None
        self.operand_ranges = {
            '-': (0, 99),  # Keep it 2 digits for better fit on small screens
            '+': (0, 99),
            '/': (0, 10),
            '*': (0, 10)
        }
        self.next()
        self.last_answer = None

    def render(self, screen: Screen):
        super().render(screen)

        if self.ready_for_next:
            # Set operand and numbers
            self.operand = choice(list(self.operand_ranges))
            min_value, max_value = self.operand_ranges[self.operand]
            self.a = randint(min_value, max_value)
            self.b = randint(min_value, max_value)

            if self.operand == '-' and (self.a - self.b) < 0:
                self.a, self.b = self.b, self.a

            if self.operand == '/' and self.b == 0:
                self.b = 1

            if self.operand == '/':
                self.a = self.a * self.b

            # Set text and display
            formula = '{} {} {}'.format(self.a, self.operand, self.b)
            self.formula = Formula(self.x, 5, text=formula, y_delta=0.1, add_bar=True)
            self.kids.add(self.formula)
            screen.add(self.formula)

            self.ready_for_next = False

        self.player.react(self.formula)

        if self.all_coords & self.player.coords:
            self.player.destruct(msg='You got CRUSHED!!')
        else:
            answer = int(eval(self.formula.text))
            for answer_text in list(self.player.kids):
                if answer_text.y <= self.formula.y + 5:
                    self.player.kids.remove(answer_text)
                    self.screen.remove(answer_text)

                    player_answer = int(answer_text.text)
                    if player_answer == answer and self.formula in self.screen:
                        self.kids.remove(self.formula)
                        self.screen.remove(self.formula)
                        for i, kid in enumerate(self.formula.kids):
                            if isinstance(kid, Bitmap):
                                self.screen.add(Explosion(kid.x, kid.y, size=kid.size * 2, on_finish=self.next))

                        self.player.scored()

    def number_pressed(self, number):
        if self.kids and self.player.alive:
            self.numbers_pressed.append(number)
            answer = int(eval(self.formula.text))
            player_answer = eval(''.join([str(n) for n in self.numbers_pressed[-len(str(answer)):]]).lstrip('0') or '0')

            if self.last_answer != player_answer and len(str(player_answer)) == len(str(answer)):
                answer = Text(self.player.x, self.player.y - 2, str(player_answer), y_delta=-1,
                              is_centered=True)
                self.screen.add(answer)
                self.player.kids.add(answer)
            self.last_answer = player_answer

    def next(self):
        if not self.kids:
            self.ready_for_next = True
            self.numbers_pressed = []
            self.last_answer = None



from games.screen import Screen
from games.objects import Circle, Square, Diamond


def test_add_and_remove():
    so = Screen()
    circle = Circle(1, 1)
    square = Square(2, 2)
    diamond = Diamond(2, 2)

    so.add(circle)
    assert len(so) == 1

    so.add(square, diamond)
    assert list(so) == [circle, square, diamond]

    so.remove(square)
    assert list(so) == [circle, diamond]

    assert square not in so
    assert circle in so

    so.reset()
    assert len(so) == 0


def test_enter_exit(screen):
    # These don't seem to work under pytest
    with screen as s:
        assert len(s.colors) == 6
        assert s.width
        assert s.height
        assert s.buffer


def test_buffer_render(screen):
    def without_distractions(buffer):
        content = []
        for y, row in enumerate(buffer):
            new_row = []
            for x, col in enumerate(row):
                if x != 0 and y != 0 and x != len(row) - 1 and y != len(buffer) - 1 and (col[0] or col[1]):
                    new_row.append(col)
            if new_row:
                content.append(new_row)
        return content

    with screen as s:
        assert s.renders == 0

        s.add(Circle(10, 10))
        s.render()

        assert s.renders == 1
        assert without_distractions(s.buffer.screen) == [
            [('O', None), ('O', None), ('O', None)],
            [('O', None), ('O', None)],
            [('O', None), ('O', None), ('O', None)]]

        s.reset()
        s.add(Circle(10, 10, color=s.COLOR_RED))
        s.render()

        assert s.renders == 1
        assert without_distractions(s.buffer.screen) == [
            [('O', 256), ('O', 256), ('O', 256)],
            [('O', 256), ('O', 256)],
            [('O', 256), ('O', 256), ('O', 256)]]



from games.controller import Controller
from games.geo_bash.scenes import ChoosePlayer, Intro, Bash
from games.objects import Triangle, Circle, Diamond, ScreenObjectGroup


class GeoBash(Controller):
    name = "Geometry Bash"

    def init(self):
        self.scenes = [ChoosePlayer, Intro, Bash]
        self.logo = ScreenObjectGroup(0, 0, objects=[
            Triangle(0, 0, color=self.screen.COLOR_BLUE),
            Circle(0, 0, color=self.screen.COLOR_RED),
            Diamond(0, 0, color=self.screen.COLOR_YELLOW)
        ])
        #: Player object that will be set by ChoosePlayer scene and then used by other scenes.
        self.player = None



from unittest.mock import Mock
from games.objects import (AbstractPlayer, Stickman, ScreenObject, Circle, Char,
                           ScreenObjectGroup, CompassionateBoss, AbstractEnemies, Bitmap, Text)


def test_player(screen):
    controller = Mock()
    player = AbstractPlayer('Max', Stickman(0, 0), controller, score_title='Wins', show_total=True)

    assert player.alive
    assert player.visible
    assert player.score == player.high_score == player.total_score

    player.scored()
    assert player.score == player.high_score == player.total_score == 1

    player.scored(points=2)
    assert player.score == player.high_score == player.total_score == 3

    assert player.screen is None
    player.render(screen)
    assert player.screen == screen
    assert screen.status['Wins'] == '3'

    player.destruct()
    assert not player.alive and not player.visible

    player.reset()
    assert player.score == 0
    assert player.high_score == player.total_score == 3

    player.scored()
    assert player.score == 1
    assert player.high_score == 3
    assert player.total_score == 4

    player.render(screen)
    assert screen.status['Wins'] == '1 | High: 3 | Total: 4'


def test_all_kids(screen):
    so = ScreenObject(1, 1)
    circle = Circle(10, 10)
    circle.add_kid(Char(10, 10, char='X'))
    so.add_kid(circle)

    assert len(so.all_kids) == 2

    list(circle.kids)[0].render(screen)
    assert so.all_coords == {(10, 10)}

    circle.render(screen)
    assert so.all_coords == {(10, 11), (9, 9), (12, 10), (8, 10), (11, 9), (10, 10), (10, 9), (9, 11), (11, 11)}


def test_is_out(screen):
    so = Bitmap(0, 0, x_delta=-1)
    assert not so.is_out

    so.render(screen)
    assert so.x == -1
    assert not so.is_out

    for i in range(so.size):
        so.render(screen)
    assert so.x == -so.size - 1
    assert so.is_out


def test_sync(screen):
    so = ScreenObject(0, 0, size=0)
    so.sync(ScreenObject(1, 1, size=1))
    assert so.x == so.y == so.size == 1


def test_reset(screen):
    so = Text(0, 0, "Hello")
    circle = Circle(1, 1)
    so.add_kid(circle)
    screen.add(so, circle)

    assert len(so.kids) == 1
    assert so in screen
    assert circle in screen

    so.render(screen)
    so.reset()

    assert len(so.kids) == 0
    assert so in screen
    assert circle not in screen


def test_add_remove_kid():
    so = ScreenObject(0, 0, size=0)
    kid = Circle(1, 1)

    so.add_kid(kid)
    assert kid in so.kids

    so.remove_kid(kid)
    assert kid not in so.kids
    assert len(so.kids) == 0

    so.remove_kid(kid)   # No op


def test_screen_object_group(screen):
    char = Char(0, 0, char='X')
    circle = Circle(0, 0)
    sog = ScreenObjectGroup(10, 10)
    sog.add(char, circle)

    screen.add(sog)
    with screen:
        screen.render()
        screen.render()

    assert sog.kids == {char, circle}

    assert sog.all_coords == {(14, 10), (11, 11), (10, 10), (11, 9), (8, 10),
                              (12, 11), (13, 9), (12, 9), (13, 11)}

    sog.y += 10
    with screen:
        screen.render()
    assert sog.all_coords == {(14, 20), (11, 21), (10, 20), (11, 19), (8, 20),
                              (12, 21), (13, 19), (12, 19), (13, 21)}


def test_compassionate_boss(screen, player):
    boss = CompassionateBoss('Max', Circle(1, 1,), player)
    screen.add(boss)
    boss.render(screen)
    assert boss.y == 1.1


def test_enemies(screen, player):
    class Enemies(AbstractEnemies):
        def create_enemy(self):
            return Circle(1, 1, x_delta=1, y_delta=2)

        def should_spawn_boss(self):
            return self.player.score == 10

        def create_boss(self):
            return Char(20, 1, char='B', y_delta=1)

    # First render creates 1 enemy
    enemies = Enemies(player)
    enemies.render(screen)
    screen.add(enemies)

    assert len(enemies.enemies) == 1
    assert len(screen) == 2  # enemies + first_enemy
    first_enemy = list(enemies.enemies)[0]

    # Next 10 renders will create max of 5 enemies and no more
    with screen:
        for i in range(10):
            screen.render()
    assert len(enemies.enemies) == 5
    assert len(screen) == 6

    assert first_enemy.x == 11
    assert first_enemy.y == 21

    # Satisfy condition to create boss
    player.score = 10
    with screen:
        screen.render()
    assert len(enemies.enemies) == 6
    assert len(screen) == 7

    # No more enemies will be created
    with screen:
        screen.render()
    assert len(enemies.enemies) == 6
    assert len(screen) == 7

    # Next 10 renders will remove all enemies as they get out of screen when player is dead
    player.alive = False
    with screen:
        for i in range(10):
            screen.render()
    print(enemies.enemies)
    assert len(enemies.enemies) == 0
    assert len(screen) == 1  # Just "enemies" object



from time import time
import curses

from games.screen import Screen, Scene
from games.objects import KeyListener, ScreenObject


class Controller(KeyListener):
    #: Name of the game
    name = None

    def __init__(self, screen: Screen):
        if not self.name:
            raise ValueError('Please set class attribute `name` to the name of the game')

        self.screen = screen
        self.reset()

    def reset(self):
        self.key_listeners = set()
        self._key_presses = set()
        self.current_index = 0
        self.current_scene = None
        self.last_key_pressed = None
        self.last_key_time = None
        self.done = False
        self.scenes = []
        self.logo = ScreenObject(0, 0)
        self.init()

    def init(self):
        """ Initialize the game with scenes and other global objects used across scenes:

                self.scenes = [SceneClass1, SceneClass2]    # Required
                self.logo = AnyScreenObjectClass()          # Optional
                self.player = ...                           # Optional

            :raises ValueError: `self.scenes` must be set by overriding this function.
        """
        raise ValueError('Please override init() to set self.scenes to a list of Scene classes')

    def set_scene(self, scene: Scene):
        """ Set the given scene as the active scene to be rendered """
        self.screen.reset()
        self.current_scene = scene
        self.key_listeners = {scene}
        self._key_presses = set()
        scene.start()

    def reset_scene(self):
        """ Restart the current scene """
        if self.current_scene:
            self.set_scene(self.current_scene.__class__(self.screen, self))

    def play(self):
        """ Handle play logic, such as key presses """
        self.screen.border.title = self.name

        if not self.current_scene:
            self.set_scene(self.scenes[self.current_index](self.screen, self))
        elif self.current_scene.done:
            self.current_index += 1
            if self.current_index >= len(self.scenes):
                self.current_index = 0
            self.set_scene(self.scenes[self.current_index](self.screen, self))

        key = self.next_key()

        if key:
            self.key_pressed(key)
        elif self.last_key_time and (time() - self.last_key_time) > 0.5:
            self.key_released()

        if key:
            if key == curses.KEY_LEFT or key == ord('h'):
                self.left_pressed()

            elif key == curses.KEY_RIGHT or key == ord('l'):
                self.right_pressed()

            elif key == curses.KEY_UP or key == ord('k'):
                self.up_pressed()

            elif key == curses.KEY_DOWN or key == ord('j'):
                self.down_pressed()

            elif key == 27 or key == ord('q'):
                self.escape_pressed()

            elif key == ord(' ') or key == ord('f'):
                self.space_pressed()

            elif key == 10:
                self.enter_pressed()

            elif key >= 48 and key <= 57:
                self.number_pressed(key - 48)

            elif key == 45:
                self.minus_pressed()

    def next_key(self):
        """ Get the next unique key from a series of presses """
        key = self.screen.key
        last_key = None
        while key > 0:  # Drain the key buffer to avoid input lag
            if key != last_key and key not in self._key_presses:
                self._key_presses.add(key)
                last_key = key
            key = self.screen.key

        return self._key_presses.pop() if self._key_presses else None

    def key_pressed(self, key):
        if key == ord('D') == self.last_key_pressed:
            self.screen.debug()

        # self.screen.debug(key=key)

        self.last_key_pressed = key
        self.last_key_time = time()

        for listener in self.key_listeners:
            listener.key_pressed(key)

    def key_released(self):
        for listener in self.key_listeners:
            listener.key_released()

    def left_pressed(self):
        for listener in self.key_listeners:
            listener.left_pressed()

    def right_pressed(self):
        for listener in self.key_listeners:
            listener.right_pressed()

    def up_pressed(self):
        for listener in self.key_listeners:
            listener.up_pressed()

    def down_pressed(self):
        for listener in self.key_listeners:
            listener.down_pressed()

    def escape_pressed(self):
        for listener in self.key_listeners:
            listener.escape_pressed()

    def space_pressed(self):
        for listener in self.key_listeners:
            listener.space_pressed()

    def enter_pressed(self):
        for listener in self.key_listeners:
            listener.enter_pressed()

    def minus_pressed(self):
        for listener in self.key_listeners:
            listener.minus_pressed()

    def number_pressed(self, number):
        for listener in self.key_listeners:
            listener.number_pressed(number)



from games.screen import Scene
from games.objects import Monologue
from games.wasp_invasion.objects import Enemies, Landscape0, Landscape1, Landscape2


class Intro(Scene):
    def init(self):
        self.intro = Monologue(self.controller.player.x, self.controller.player.y - 3,
                               on_finish=self.next,
                               texts=["Hi, I am {}.".format(self.controller.player.name),
                                      "Wasp kaijus have invaded Earth!!",
                                      "We don't know why they are here,",
                                      "but they have taken control of all wasps",
                                      "and started killing everyone.",
                                      "Luckily, I got my trusty flamethrower with me.",
                                      "Oh no! Here they come. GET READY!!",
                                      "Move around using left/right keys",
                                      "and control the flame using up/down keys."])
        self.controller.player.reset()
        self.controller.player.active = False

    def start(self):
        self.screen.add(self.intro, self.controller.player)

    def key_pressed(self, key):
        self.next()


class Survive(Scene):
    def init(self):
        self.enemies = Enemies(self.controller.player, max_enemies=int(self.screen.width / 10))
        self.landscapes = (Landscape0(0, 0), Landscape1(0, 0), Landscape2(0, 0))
        self.player.reset()

    def start(self):
        self.screen.add(*self.landscapes, self.player, self.enemies)

    def escape_pressed(self):
        self.controller.done = True



from games.controller import Controller
from games.objects import One, Two, Plus, ScreenObjectGroup
from games.number_crush.scenes import Crush, Intro
from games.number_crush.objects import Player


class NumberCrush(Controller):
    name = "Number Crush"

    def init(self):
        self.scenes = [Intro, Crush]
        self.logo = ScreenObjectGroup(0, 0, objects=[
            One(0, 0),
            Plus(0, 0),
            Two(0, 0)
        ])
        self.player = Player(self)



from games.controller import Controller
from games.wasp_invasion.scenes import Intro, Survive
from games.wasp_invasion.objects import Player
from games.objects import WaspKaiju, Stickman


class WaspInvasion(Controller):
    name = "Wasp Invasion"

    def init(self):
        self.scenes = [Intro, Survive]
        self.logo = WaspKaiju(0, 0, color=self.screen.COLOR_YELLOW)
        self.player = Player('Jon', Stickman(self.screen.width / 2, self.screen.height - 2), self)
        self.player.controller = self



from games.controller import Controller
from games.last_survivor.scenes import Intro, Survive
from games.last_survivor.objects import Player
from games.objects import Zombie, Char


class LastSurvivor(Controller):
    name = "THE LAST SURVIVOR!!"

    def init(self):
        self.scenes = [Intro, Survive]
        self.logo = Zombie(0, 0, color=self.screen.COLOR_GREEN)
        self.player = Player('Jon', Char(self.screen.width / 2, self.screen.height / 2, char='☻'), self)
        self.player.controller = self



from random import randint, random, choice

from games.screen import Screen
from games.objects import (Wasp, Explosion, Projectile, Monologue,
                           Stickman, AbstractPlayer, AbstractEnemies, CompassionateBoss,
                           WaspKaiju, DyingWaspKaiju, Char, Landscape)


class Player(AbstractPlayer):
    def __init__(self, *args, **kwargs):
        self.gas_limit = 100
        self.left_deltas = (-2, 0, '⇚')
        self.upleft_deltas = (-2, -1, '⇖')
        self.right_deltas = (2, 0, '⇛')
        self.upright_deltas = (2, -1, '⇗')
        self.projectile_deltas = self.right_deltas

        super().__init__(*args, score_title='Killed', max_hp=100, **kwargs)

    def reset(self):
        super().reset()
        self.sync(self.shape)

        self.gas = self.gas_limit
        self.flame_on = True

        if self.screen:
            self.x = self.screen.width / 2
            self.y = self.screen.height - 2

        self.flamethrower = Char(self.x, self.y, char=None)

    def render(self, screen: Screen):
        super().render(screen)

        if self.active:
            screen.border.set_levels(self.hp / self.max_hp, self.gas / self.gas_limit)

            if self.is_hit:
                self.color = screen.COLOR_YELLOW
                self.is_hit = False
            else:
                self.color = None

            if self.y_delta:
                # Jumps up
                if self.y_delta < 0:
                    self.y_delta *= 0.9

                    if self.y_delta > -0.2:
                        self.y_delta *= -1

                # Comes back down
                else:
                    if self.y_delta < 0.5:
                        self.y_delta *= 1.1

                    if self.y >= self.screen.height - 2.5:
                        self.y_delta = 0

            x_delta, y_delta, char = self.projectile_deltas
            self.flamethrower.x = self.x + x_delta - 0.5
            self.flamethrower.y = self.y + y_delta - 0.5

            if self.flame_on:
                self.flamethrower.char = char
                self.flamethrower.color = choice([screen.COLOR_RED, screen.COLOR_YELLOW])

                if self.flamethrower not in screen:
                    screen.add(self.flamethrower)

                if self.gas > 0:
                    for flame_size in range(min(20, int(screen.width / 9))):
                        explosion = Explosion(self.x, self.y, size=flame_size, parent=self)
                        projectile = Projectile(self.x, self.y, shape=None, parent=self,
                                                x_delta=x_delta * (1 + flame_size / 20),
                                                y_delta=y_delta, color=screen.COLOR_YELLOW,
                                                explode_after_renders=flame_size,
                                                explosion=explosion)
                        self.kids.add(projectile)
                        self.screen.add(projectile)

                if self.gas > 0:
                    self.gas -= 0.01
                elif self.gas != -1:
                    screen.add(Monologue(self.x, self.y - 2, texts=['Out of gas!!!',
                                                                    'Look for green gas refills']))
                    self.gas = -1
            else:
                self.flamethrower.char = '⇓'
                self.flamethrower.color = None
                self.flamethrower.y = self.y + 0.5

    def destruct(self):
        super().destruct(msg='You got STUNG!!', explosion_size=30)
        self.screen.add(Stickman(self.shape.x, self.shape.y, color=self.screen.COLOR_YELLOW, y_delta=-0.1))

    def left_pressed(self):
        if self.alive:
            if self.x > self.size * 5:
                self.x -= 1

            if self.projectile_deltas in (self.upright_deltas, self.upleft_deltas):
                self.projectile_deltas = self.upleft_deltas
            else:
                self.projectile_deltas = self.left_deltas

    def right_pressed(self):
        if self.alive:
            if self.x < self.screen.width - self.size * 5:
                self.x += 1

            if self.projectile_deltas in (self.upleft_deltas, self.upright_deltas):
                self.projectile_deltas = self.upright_deltas
            else:
                self.projectile_deltas = self.right_deltas

    def up_pressed(self):
        if self.alive:
            if not self.flame_on:
                self.flame_on = True
            elif self.projectile_deltas == self.right_deltas:
                self.projectile_deltas = self.upright_deltas
            elif self.projectile_deltas == self.left_deltas:
                self.projectile_deltas = self.upleft_deltas
            elif not self.y_delta:
                self.y_delta = -1

    def down_pressed(self):
        if self.alive and self.y_delta and self.y_delta < 0:
            self.y_delta *= -1

        if self.projectile_deltas in (self.left_deltas, self.right_deltas):
            self.flame_on = False
        elif self.projectile_deltas == self.upleft_deltas:
            self.projectile_deltas = self.left_deltas
        else:
            self.projectile_deltas = self.right_deltas

    def space_pressed(self):
        pass


class Enemies(AbstractEnemies):
    def create_enemy(self):
        if random() < 0.5:
            x = choice([0, self.screen.width])
            y = randint(0, self.screen.height)
        else:
            y = 0
            x = randint(0, self.screen.width)

        speed = random() * self.player.score / 10000 + 0.2
        x_sign = (1 if x < self.player.x else -1) * random()
        y_sign = (1 if y < self.player.y else -1) * random()

        return Wasp(x, y, x_delta=x_sign * speed, y_delta=y_sign * speed, random_start=True,
                    flip=x_sign > 0, color=self.screen.COLOR_YELLOW)

    def on_death(self, enemy):
        enemy_class = DyingWaspKaiju if isinstance(enemy, CompassionateBoss) else enemy.__class__
        wasp = enemy_class(enemy.x, enemy.y, color=self.screen.COLOR_RED, remove_after_animation=True,
                           flip=getattr(enemy, 'flip', False))
        self.screen.add(wasp)

    def should_spawn_boss(self):
        return self.player.score and self.player.score % 50 == 0

    def create_boss(self):
        if random() < 0.5:
            x_delta = -0.1
            x = self.screen.width
        else:
            x_delta = 0.1
            x = 0

        return CompassionateBoss('Kaiju',
                                 WaspKaiju(x, self.screen.height - 10,
                                           x_delta=x_delta, y_delta=0.01,
                                           color=self.screen.COLOR_YELLOW, random_start=True),
                                 self.player,
                                 hp=self.player.score/25)


class Landscape0(Landscape):
    bitmap = """
          S
"""


class Landscape1(Landscape):
    move_speed = 0.1
    bitmap = """
 
 
   T    T           T        T       T      T           T              T   T            T      T   T
"""  # noqa


class Landscape2(Landscape):
    move_speed = 0.2
    bitmap = """
 
 
 
TT       T     T    T     T     T  T    T         T   T     T   TT   T   T T TTT   T    T    T     T TT
"""  # noqa



import setuptools


setuptools.setup(
    name='console-games',
    version='0.1.28',

    author='Max Zheng',
    author_email='maxzheng.os@gmail.com',

    description='Fun games for macOS/Linux console!',
    long_description=open('README.rst').read(),

    url='https://github.com/maxzheng/console-games',

    install_requires=open('requirements.txt').read(),

    license='MIT',

    packages=setuptools.find_packages(),
    include_package_data=True,

    python_requires='>=3.5',
    setup_requires=['setuptools-git', 'wheel'],

    entry_points={
       'console_scripts': [
           'play = games.scripts:main',
       ],
    },

    # Standard classifiers at https://pypi.org/classifiers/
    classifiers=[
      'Development Status :: 5 - Production/Stable',

      'Intended Audience :: Developers',

      'License :: OSI Approved :: MIT License',

      'Programming Language :: Python :: 3',
    ],

    keywords='games terminal console geometry bash number crush math training',
)



from games.controller import Controller
from games.objects import Choice, Logo
from games.screen import Scene
from games.geo_bash import GeoBash
from games.number_crush import NumberCrush
from games.last_survivor import LastSurvivor
from games.wasp_invasion import WaspInvasion


class Chooser(Controller):
    name = "Choose a Game"

    def __init__(self, *args, game_filter=None, **kwargs):
        super().__init__(*args, **kwargs)

        #: Filter games that matches the name
        self.game_filter = game_filter

    def init(self):
        self.scenes = [Choose]


class Choose(Scene):
    def init(self):
        self.games = [GeoBash(self.screen), NumberCrush(self.screen), LastSurvivor(self.screen),
                      WaspInvasion(self.screen)]
        if self.controller.game_filter:
            self.games = list(filter(lambda g: self.controller.game_filter.lower() in g.name.lower(), self.games))
        self.game_logos = [Logo(0, 0, g.name, g.logo) for g in self.games]
        self.game_choice = Choice(x=self.screen.width / 2, y=int(self.screen.height / 2),
                                  choices=self.game_logos, on_select=self.play,
                                  current=getattr(self.controller, '_last_game_index', None))

        if not len(self.games):
            exit('No games matching name')
        elif len(self.games) == 1:
            self.play(self.game_logos[0])

    def start(self):
        self.screen.reset(border=True)
        self.screen.add(self.game_choice)

    def escape_pressed(self):
        self.controller.done = True

    def play(self, game_logo):
        game = self.games[self.game_logos.index(game_logo)]
        game.reset()

        self.screen.controller = game
        self.screen.reset()

        while not game.done:
            self.screen.render()
            game.play()

        if len(self.games) == 1:
            self.controller.done = True
        else:
            self.screen.controller = self.controller
            self.controller._last_game_index = self.game_logos.index(game_logo)
            self.controller.reset_scene()
